import internetarchive as ia
import requests

from youtube_archivist.base import JsonArchivist, LOG


class IAArchivist(JsonArchivist):
    VALID_FORMATS = ['MPEG2', "Ogg Video", "512Kb MPEG4", 'h.264']

    def archive(self, url):
        try:
            self.archive_item(url)
        except:
            self.archive_collection(url)

    def bootstrap_from_url(self, url):
        self.log.info(f"Bootstrapping database from: {url}")
        cache = requests.get(url).json()
        self.db.update(cache)
        self.db.store()

    def archive_item(self, item_id):
        item = ia.get_item(item_id)
        meta = requests.get(item.urls.metadata).json()
        tags = []
        if meta["metadata"].get('subject'):
            if isinstance(meta["metadata"]["subject"], str):
                tags += meta["metadata"]["subject"].split(";")
            if isinstance(meta["metadata"]["subject"], list):
                tags += meta["metadata"]["subject"]
        title = meta["metadata"]["title"]
        if isinstance(title, list):
            title = title[0]
        movie = {
            "collection": meta["metadata"]["collection"],
            "tags": tags,
            "streams": [],
            "title": title,
            "duration": meta["metadata"].get('runtime'),
            "images": []
        }
        for f in meta["files"]:
            if f["format"] in self.VALID_FORMATS:
                stream = item.urls.download + "/" + f["name"]
                movie["streams"].append(stream)
            if f["format"] in ["PNG"]:
                movie["images"] += [item.urls.download + "/" + f["name"]]
        if not movie["streams"]:
            return
        if any(k.lower() in movie["title"].lower() for k in
               self.blacklisted_kwords):
            return
        if any(k.lower() not in movie["title"].lower() for k in
               self.required_kwords):
            return
        #if movie["duration"] < self.min_duration:
        #    return
        self.log.info("Parsing video " + movie["title"])
        self.db[item_id] = movie
        self.db.store()

    def archive_collection(self, collection_name):
        session = ia.ArchiveSession()
        entries = list(ia.Search(session, 'collection:' + collection_name))
        for idx, entry in enumerate(entries):
            item_id = entry['identifier']
            if item_id not in self.db:
                self.archive_item(item_id)



from logging import getLogger
from json_database import JsonStorageXDG


LOG = getLogger("youtube_archivist")
LOG.setLevel("DEBUG")


class JsonArchivist:
    def __init__(self, db_name=None, required_kwords=None, blacklisted_kwords=None, min_duration=-1, logger=LOG):
        self.required_kwords = required_kwords or []
        self.blacklisted_kwords = blacklisted_kwords or []
        self.db = JsonStorageXDG(db_name, subfolder="youtube_archivist")
        self.min_duration = min_duration
        self.log = logger

    @property
    def video_urls(self):
        return list(self.db.keys())

    def archive(self, url):
        raise NotImplementedError

    def sorted_entries(self):
        return sorted([e for e in self.db.values()],
                      key=lambda k: k.get("upload_ts"),
                      reverse=True)

    # DB interaction
    def remove_unavailable(self):
        pass

    def remove_keyword(self, kwords=None):
        kwords = kwords or self.blacklisted_kwords
        bad_urls = []
        for url, entry in self.db.items():
            name = entry["title"]
            if any([k.lower() in name.lower() for k in kwords]):
                bad_urls.append(url)
        for url in bad_urls:
            self.db.pop(url)
            LOG.info("Removed entry: " + url)
        self.db.store()

    def remove_missing(self, kwords):
        bad_urls = []
        for url, entry in self.db.items():
            if any([not entry.get(k) for k in kwords]):
                bad_urls.append(url)
        for url in bad_urls:
            self.db.pop(url)
            LOG.info("Removed entry: " + url)
        self.db.store()

    def remove_below_duration(self, minutes=30):
        bad_urls = []
        for url, entry in self.db.items():
            dur = entry.get("duration") or 0
            if dur <= minutes * 60:
                bad_urls.append(url)
        for url in bad_urls:
            self.db.pop(url)
            LOG.info("Removed entry: " + url)
        self.db.store()



from tutubo.models import Channel, Playlist, Video
from youtube_archivist.exceptions import VideoUnavailable
from youtube_archivist.base import JsonArchivist, LOG
from queue import Queue
from threading import Thread, Event
import time
import requests


class YoutubeMonitor(Thread):
    queue = Queue()

    def __init__(self, db_name=None, required_kwords=None, blacklisted_kwords=None, min_duration=-1, logger=LOG):
        super().__init__()
        self.archive = YoutubeArchivist(db_name, required_kwords, blacklisted_kwords, min_duration)
        self.monitoring = Event()
        self.repeat_list = {}
        self.log = logger

    @property
    def db(self):
        return self.archive.db

    def sorted_entries(self):
        return sorted([e for e in self.db.values()],
                      key=lambda k: k.get("upload_ts"),
                      reverse=True)

    def bootstrap_from_url(self, url):
        if not self.archive.db:
            self.log.info(f"Bootstrapping database from: {url}")
            cache = requests.get(url).json()
            self.archive.db.update(cache)
            self.archive.db.store()

    def _index_url(self, url):
        # index url contents
        try:
            if url in self.repeat_list:
                # handle repeating checks
                if time.time() - self.repeat_list[url] < 30:
                    # exception caught below
                    # its basically an interrupt
                    raise RuntimeError("Wait before syncing again")
                self.repeat_list[url] = time.time()
            self.archive.archive(url)
        except RuntimeError:
            pass

    def run(self):
        self.monitoring.set()
        self.log.info(f"Started monitoring: {self.archive.db.name}")

        # remove any deleted videos
        try:
            self.archive.remove_unavailable()
        except Exception as e:
            self.log.exception(e)

        while self.monitoring.is_set():
            url = self.queue.get()

            # index url contents
            try:
                self._index_url(url)
            except Exception as e:
                self.log.exception(e)
                pass

            # sleep for a while
            time.sleep(120)

            # keep monitoring url
            if url in self.repeat_list:
                self.queue.put(url)  # sync this url again

    def sync(self, url):
        self.queue.put(url)

    def monitor(self, url):
        if url not in self.repeat_list:
            self.repeat_list[url] = 0
        self.sync(url)

    def stop(self):
        self.monitoring.clear()


class YoutubeArchivist(JsonArchivist):

    def archive(self, url):
        if "/channel/" in url or "/c/" in url:
            return self.archive_channel(url)
        if "/watch" in url:
            return self.archive_video(url)
        if "/playlist" in url:
            return self.archive_playlist(url)
        return self.archive_channel(url)

    def archive_video(self, url, extra_data=None):
        urls = self.video_urls
        if isinstance(url, str):
            video = Video(url)
        else:
            video = url
        if video.watch_url not in urls:
            self.log.debug(f"Archiving video: {url}")
            try:
                if any(k.lower() in video.title.lower() for k in
                       self.blacklisted_kwords):
                    return
                if any(k.lower() not in video.title.lower() for k in
                       self.required_kwords):
                    return
                if video.length < self.min_duration:
                    return
                self.log.info("Parsing video " + video.title)
                self._update_video(video, extra_data)
            except:
                # accessing the title property might cause a 404 if
                # video was removed
                raise
        del video

    def archive_playlist(self, url):
        self.log.debug(f"Archiving playlist: {url}")
        c = Playlist(url)
        try:
            meta = {"playlist": c.title}
        except:
            meta = {}
        for video in c.videos:
            self.archive_video(video, meta)

    def archive_channel(self, url):
        self.log.debug(f"Archiving channel: {url}")
        c = Channel(url)
        for video in c.videos:
            meta = {}
            self.archive_video(video, meta)

    def archive_channel_playlists(self, url):
        self.log.debug(f"Archiving  channel playlists: {url}")
        c = Channel(url)
        for pl in c.playlists:
            try:
                meta = {"playlist": pl.title}
            except:
                meta = {}
            for video in pl.videos:
                try:
                    self.archive_video(video, meta)
                except VideoUnavailable:
                    continue

    def _update_video(self, entry, extra_data=None):
        if not entry:
            return
        url = entry.watch_url
        # format data for saving
        entry = {
            "author": entry.author,
            "title": entry.title,
            "url": entry.watch_url,
            "duration": entry.length,
            "upload_ts": entry.publish_date.timestamp(),
            "tags": entry.keywords,
            'thumbnail': entry.thumbnail_url
        }
        if extra_data:
            entry.update(extra_data)
        self.db[url] = entry
        self.db.store()

    # DB interaction
    def remove_unavailable(self):
        to_remove = []
        for url, entry in self.db.items():
            vid = Video(url)
            try:
                vid_data = {
                    "videoId": vid.video_id,
                    "url": vid.watch_url,
                    "image": vid.thumbnail_url,
                    "title": vid.title
                }
            except VideoUnavailable:
                to_remove.append(url)
        for url in to_remove:
            self.db.pop(url)
            self.log.info("Removed entry: " + url)
        self.db.store()


from setuptools import setup

setup(
    name='youtube_archivist',
    version='0.0.3a1',
    packages=['youtube_archivist'],
    url='https://github.com/OpenJarbas/youtube_archivist',
    license='apache2',
    install_requires=["json_database>=0.3.0", "tutubo", "internetarchive"],
    author='jarbasai',
    author_email='jarbasai@mailfence.com',
    description='youtube indexer - keep track of your favorite channels!'
)
